#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h> 	
#include <stdio.h>

#include "serial.h"
#include "spi.h"
#include "ordonne.h"

#define	DEFAULT_BRIGHTNESS	128
#define NB_TICK		312
#define MAX_PROCS 3

#define SAVE_REGISTER()  \
	asm volatile ( \
		"push	r0				\n\t" \
		"in		r0, __SREG__	\n\t" \
		"push	r0				\n\t" \
		"push	r1				\n\t" \
		"push	r2				\n\t" \
		"push	r3				\n\t" \
		"push	r4				\n\t" \
		"push	r5				\n\t" \
		"push	r6				\n\t" \
		"push	r7				\n\t" \
		"push	r8				\n\t" \
		"push	r9				\n\t" \
		"push	r10				\n\t" \
		"push	r11				\n\t" \
		"push	r12				\n\t" \
		"push	r13				\n\t" \
		"push	r14				\n\t" \
		"push	r15				\n\t" \
		"push	r16				\n\t" \
		"push	r17				\n\t" \
		"push	r18				\n\t" \
		"push	r19				\n\t" \
		"push	r20				\n\t" \
		"push	r21				\n\t" \
		"push	r22				\n\t" \
		"push	r23				\n\t" \
		"push	r24				\n\t" \
		"push	r25				\n\t" \
		"push	r26				\n\t" \
		"push	r27				\n\t" \
		"push	r28				\n\t" \
		"push	r29				\n\t" \
		"push	r30				\n\t" \
		"push	r31				\n\t" \
	);

#define RESTORE_REGISTER() \
	asm volatile (	\
 		"pop r31 				\n\t" \
 		"pop r30 				\n\t" \
 		"pop r29 				\n\t" \
 		"pop r28 				\n\t" \
 		"pop r27 				\n\t" \
 		"pop r26 				\n\t" \
 		"pop r25 				\n\t" \
 		"pop r24 				\n\t" \
 		"pop r23 				\n\t" \
 		"pop r22 				\n\t" \
 		"pop r21 				\n\t" \
 		"pop r20 				\n\t" \
 		"pop r19 				\n\t" \
 		"pop r18 				\n\t" \
 		"pop r17 				\n\t" \
 		"pop r16 				\n\t" \
 		"pop r15 				\n\t" \
 		"pop r14 				\n\t" \
 		"pop r13 				\n\t" \
 		"pop r12 				\n\t" \
 		"pop r11 				\n\t" \
 		"pop r10 				\n\t" \
 		"pop r9 				\n\t" \
 		"pop r8 				\n\t" \
 		"pop r7 				\n\t" \
 		"pop r6 				\n\t" \
 		"pop r5 				\n\t" \
 		"pop r4 				\n\t" \
 		"pop r3 				\n\t" \
 		"pop r2 				\n\t" \
 		"pop r1 				\n\t" \
 		"pop r0 				\n\t" \
 		"out __SREG__, r0 		\n\t" \
 		"pop r0 				\n\t" \
	);

uint8_t pid=0;

//déclaration de la structure contenant les processus
typedef struct scheduler_t {
  uint16_t pile; //position courante dans la fonction associé au processus
  void (*fonction) (void); // fonction associée au processus
}scheduler_t;

//remplissage de la structure contenant les processus
scheduler_t processus[MAX_PROCS]={
  {0x500,task_7segments},
  {0x700,task_matriceLEDs},
  {0x600,task_CommunicationSPI},
  };

//initialisation du timer qui permettra la rotation des processus
void init_timer()
{
	TCCR1B |= _BV(WGM12); // CTC mode with value in OCR1A 
	TCCR1B |= _BV(CS12);  // CS12 = 1; CS11 = 0; CS10 =1 => CLK/1024 prescaler
	TCCR1B |= _BV(CS10);
	OCR1A   = NB_TICK;
	TIMSK1 |= _BV(OCIE1A);
}

//Routine d'interruption du timer
ISR(TIMER1_COMPA_vect, ISR_NAKED)
{
    SAVE_REGISTER(); //On sauvegarde la pile, qui contient le contexte d'exécution du processus 
    processus[pid].pile=SP; //ainsi qu'où était le pointeur de pile, afin de reprendre où nous en étions lorsque ce processus sera exécuté à nouveau	
	
    pid=(pid+1)%MAX_PROCS; //on passe au processus suivant
    
    SP = processus[pid].pile; // on récupère où en était le pointeur de pile à la précédente exécution
    RESTORE_REGISTER(); //et on recharge la pile

    asm volatile("reti"); // on autorise à nouveau les interruptions
}

//fonction permettant la préparation des processus avant la première exécution
void init_proc(int i)
{
    uint16_t old = SP; // on sauvegarde où était le pointeur de pile avant l'initialisation du processus

    SP = processus[i].pile; // on va à l'emplacement que l'on à initialement choisi pour notre processus lorsque l'on a rempli la structure
    int adresse=(int)processus[i].fonction; //on recupere l'adresse où a été mis la fonction de notre processus pour la pousser sur la pile
    asm volatile("push %0" : : "r" (adresse & 0x00ff) ); //d'abord les 8 bits de poids faibles
    asm volatile("push %0" : : "r" ((adresse & 0xff00)>>8)); //puis les 8 bits de poids forts que l'on décale donc de 8 bits
    
    SAVE_REGISTER(); // on sauvegarde la pile
    processus[i].pile = SP; //ainsi qu'où était le pointeur de pile, le processus est prêt pour sa première utilisation

    SP = old; //on remet le pointeur de pile où il était avant l'initialisation du processus
}

void task_7segments()
{
	uint8_t display7Segments1[] = {5, 'A', 4, 'B', 3, 'C', 2, 'D', 1, 'E', 0, 'F'};
	uint8_t display7Segments2[] = {0, 'A', 1, 'B', 2, 'C', 3, 'D', 4, 'E', 5, 'F'};
    /*Permet l'envoie d'un tableau sur les afficheurs 7 segments*/
    spi_select(&SPI_PORT_DISPLAY,SPI_BIT_DISPLAY);//sélection du uC controlant les afficheurs 7 segments
    for(int i=0;i<12;i++) //parcours du tableau pour l'envoie des valeurs
    {
      if(rotationmessage)
      {
      	// tant que la réponse est différente de 255 et donc que le µC est occupé, l'octet est renvoyé (cela provoque une saturation...)
      	while( (reponse = spi_exch(display7Segments1[i])) != 255); 
      }
      else
      {
      	// idem
      	while( (reponse = spi_exch(display7Segments2[i])) != 255);
      }
    }
    spi_unselect(&SPI_PORT_DISPLAY,SPI_BIT_DISPLAY); //déselection du uC
}

void task_matriceLEDs()
{
	uint8_t matrice[] = {'B', 'O', 'N', 'J', 'O', 'U', 'R'};
      //Permet l'envoie d'un message sur la matrice de LEDs
    spi_select(&SPI_PORT_MATRIX,SPI_BIT_MATRIX);//sélection du uC controlant la matrice de LEDs
    _delay_ms(5);
    for(int i=0;i<7;i++) //parcours du tableau pour l'envoie du message
    {
      spi_exch(matrice[i]);
      _delay_ms(800); // temps permettant d'avoir le temps de lire correctement l'enchainement des caractères
    }
    spi_unselect(&SPI_PORT_MATRIX,SPI_BIT_MATRIX); //déselection du uC
    _delay_ms(150);
}

void task_CommunicationSPI()
{
	cli(); //blocage des interruptions
	
	//cette tâche permettra d'effectuer l'envoi d'une commande (aux afficheurs 7 segments ou à la matrice de LEDs) sans être préempté
	
	sei(); //réautorisation des interruptions
}

int main(void)
{
	spi_init(1);
	_delay_ms(1000);

	spi_initcs(&SPI_DDR_DISPLAY,SPI_BIT_DISPLAY);
	spi_initcs(&SPI_DDR_MATRIX,SPI_BIT_MATRIX);
	spi_unselect(&SPI_PORT_DISPLAY,SPI_BIT_DISPLAY);
	spi_unselect(&SPI_PORT_MATRIX,SPI_BIT_MATRIX);
	
	init_timer(); //initialisation du timer
	for(int i=0;i<MAX_PROCS;i++) init_proc(i); // initialisation des processus
	sei(); //autorisation des interruptions
	processus[0].fonction(); // lancement manuel du premier processus
	  
	return 0;
}
